import os, posixpath, logging, glob, fnmatch

from concurrent.futures import ThreadPoolExecutor

from sonador.remote import sonador_datacollection_list, sonador_dataobject_details, sonador_dataobject_schema_display, \
	fetch_sonador_dataobject
from sonador.servers import SonadorImagingServerCollection, DicomImagingModalityCollection

logger = logging.getLogger(__name__)


def imageserver_datacollection_list(sonador_server, args, resource_stub, datamodelcollection_class, 
		filters=None, imageserver_datamodel_class=SonadorImagingServerCollection.model, iserver=None, filter=None,
		verify=False):
	'''	Retrieve and list a data collection associated with an imaging server

		return Data Collection
	'''
	filters = filters or {}

	# Retrieve imaging server instance
	iserver = iserver or fetch_sonador_dataobject(sonador_server, imageserver_datamodel_class, args.server,
		verify=verify)
	iserver_resourcebase = posixpath.join(iserver.fetch_endpoint, args.server, resource_stub)

	# List details of associated modalities
	dcollection = sonador_datacollection_list(sonador_server, args.output_dest, datamodelcollection_class,
		output_type=args.output_type, verify=args.verifyssl, filters=filters, 
		data_collection_endpoint=iserver_resourcebase, pacs=iserver)
	args.output_dest.write('\n')

	return dcollection


def imageserver_dataobject_details(sonador_server, args, resource_stub, datamodel_class, uidlist, detailheader_template,
		imageserver_datamodel_class=SonadorImagingServerCollection.model, iserver=None, verify=False):
	'''	Retrieve and display the detailed listings for Sonador image servers and related data models.
	'''
	# Retrieve imaging server instance
	iserver = iserver or fetch_sonador_dataobject(sonador_server, imageserver_datamodel_class, args.server,
		verify=verify)
	iserver_resourcebase = posixpath.join(iserver.fetch_endpoint, args.server, resource_stub)

	if args.include_schema:
		sonador_dataobject_schema_display(
			sonador_server, args.output_dest, datamodel_class, 
			verify=args.verifyssl, data_collection_endpoint=iserver_resourcebase)

	for uid in uidlist:
		logger.info(detailheader_template % (iserver.server_label, uid))
		dobject = sonador_dataobject_details(sonador_server, args.output_dest, datamodel_class, uid,
			verify=args.verifyssl, dataobject_endpoint=posixpath.join(iserver_resourcebase, uid))
		args.output_dest.write('\n')


def imageserver_upload_folder(iserver, folders, threads=4, verify=False, 
		dcm_extensions=['*.dcm', '*.DCM', '*.DICOM', '*.dicom', 'IM*']):
	'''	Scan folders and upload all DICOM images to the provided imaging servers

		@input iserver (SonadorImagingServer instance): Imaging server to which the
			images should be uploaded.
	'''
	# Create thread pool
	tpool = ThreadPoolExecutor(max_workers=threads)

	for froot in folders:
		logger.info('Scan %s for images to upload to server %s' % (froot, iserver.server_label))

		# Walk through folders and locate images for upload
		for croot, cfolders, cfiles in os.walk(froot):

			# Find all DICOM files and variations inside of the directory
			dcmfiles = []
			for ext in dcm_extensions:
				dcmfiles.extend(fnmatch.filter(cfiles, ext))

			# Upload files (in parallel) to the image server
			if len(dcmfiles):
				logger.debug('Found %s files in folder %s, begin upload' % (len(dcmfiles), croot))

				def upload_dcmimages(iname):
					'''	Upload the provided image to Orthanc
					'''
					ipath = os.path.join(croot, iname)
					with open(ipath, 'rb') as img:
						logger.debug('Upload image %s to server %s' % (iname, iserver.pk))

						# Upload image to PACS imaging server
						r = iserver.upload_image(img)
						return r.ok

				uresults = sum(tpool.map(upload_dcmimages, dcmfiles))
				logger.info('Transfer results (%s): %s images uploaded successfully' % (croot, uresults))
