import six, os, logging, argparse
from collections.abc import Iterable

from client.utils.urls import validate_url
from client.utils.logs import LOGGING_LEVELS

from sonador.validate import argparse_type_directory, argparse_keyval, validate_sonador_connection_args
from sonador.parser import add_arguments_for_api_connection, add_arguments_for_verify_ssl, \
	add_arguments_for_logging_options, output_operation_options, \
	output_destination_options, SONADOR_COMMAND_LIST, SONADOR_COMMAND_DETAILS, \
	OUTPUT_TYPE_TABULATE, OUTPUT_TYPE_CSV

from .apisettings import SONADOR_COMMAND_TOKEN, SONADOR_COMMAND_MODALITY, SONADOR_COMMAND_DICOMWEB, \
	SONADOR_COMMAND_SUPPORTED_MODALITY, SONADOR_COMMAND_SUPPORTED_DICOMWEB, SONADOR_COMMAND_DICOMWEB_REMOTECMD_SUPPORTED, \
	SONADOR_COMMAND_REMOTE, SONADOR_COMMAND_DICOMWEB_REMOTE_SUPPORTED, \
	SONADOR_COMMAND_IMAGE, SONADOR_COMMAND_IMAGE_UPLOAD, SONADOR_COMMAND_SUPPORTED_IMAGE, \
	SONADOR_COMMAND_QUERY, SONADOR_COMMAND_QUERY_FETCH, \
	SONADOR_COMMAND_IMAGE_SERVER, SONADOR_COMMAND_SUPPORTED_IMAGE_SERVER, SONADOR_SUPPORTED_COMMANDS, \
	SONADOR_COMMAND_PATIENT, SONADOR_COMMAND_PATIENT_SUPPORTED, \
	SONADOR_COMMAND_STUDY, SONADOR_COMMAND_STUDY_SUPPORTED, \
	SONADOR_COMMAND_JOB, SONADOR_COMMAND_PAUSE, SONADOR_COMMAND_RESUME, SONADOR_COMMAND_CANCEL, \
	SONADOR_COMMAND_RESUBMIT, SONAOD_COMMAND_SUPPORTED_JOB

logger = logging.getLogger(__name__)



def validate_args(args):
	'''	Ensure user-provided arguments meet the requirements to interface with the API

		@input args: arguments to be validated
		@returns int: exit code
	'''
	exitcode = validate_sonador_connection_args(args, 0)

	# Verify access ID and secret key
	if not args.apitoken:

		if args.accessid is None:
			logger.error(six.text_type('The import client requires a Sonador Access ID. See --help for details.'))
			exitcode = 1
		if args.secretkey is None:
			logger.error(six.text_type('The import client requires a Sonador Secret Key. See --help for details.'))
			exitcode = 1

	# Verify endpoint value and structure
	if not args.endpoint:
		logger.error(six.text_type('A Sonador endpoint is required.'))
		exitcode = 1

	# Verify Sonador URL endpoint
	if args.endpoint:

		try: validate_url(args.endpoint)
		except ValueError as err:
			logger.error(
				six.text_type('Malformed endpoint URL "%s", a valid http URL is required. Example: http://domain.com:port'
					% args.endpoint))
			exitcode = 1

	# Ensure that an operation was specified
	if not getattr(args, 'command', None):
		logger.error(six.text_type('Unsupported command: %s. Supported: %s.'
			% (getattr(args, 'command', None) or '(null)', ', '.join(SONADOR_SUPPORTED_COMMANDS))))
		exitcode = 1

	# Imaging Servers API
	if getattr(args, 'command', None) == SONADOR_COMMAND_IMAGE_SERVER:
		exitcode = validate_imageserver_args(args, exitcode)

	return exitcode


def validate_imageserver_args(args, exitcode):
	'''	Ensure user-provided arguments for the Imaging Server API are sane.

		@input args: arguments to be validated
		@returns int: exit code
	'''
	if not getattr(args, 'imageserver_command', None):
		logger.error('Please specify an image server command to execute. Supported: %s'
			% ', '.join(SONADOR_COMMAND_SUPPORTED_IMAGE_SERVER))
		exitcode = 1

	if getattr(args, 'imageserver_command', None) == SONADOR_COMMAND_DETAILS \
		and not getattr(args, 'servers', None):
		logger.error('Please specify the server instances for which you wish to retrieve the details.')
		exitcode = 1

	if getattr(args, 'imageserver_command', None) in (
		SONADOR_COMMAND_MODALITY, SONADOR_COMMAND_DICOMWEB, SONADOR_COMMAND_IMAGE, SONADOR_COMMAND_JOB):

		# Ensure that a server argument has been provided
		if not getattr(args, 'server', None):
			logger.error('Invalid PACS imaging server, please provide a valid imaging server ID. See --help '
				+ 'for additional information.')
			exitcode = 1

	# Validate options for specific interfaces
	if getattr(args, 'imageserver_command', None) == SONADOR_COMMAND_MODALITY:
		exitcode = validate_imageserver_dicom_modality_args(args, exitcode)
	elif getattr(args, 'imageserver_command', None) == SONADOR_COMMAND_DICOMWEB:
		exitcode = validate_imageserver_dicomweb_args(args, exitcode)
	elif getattr(args, 'imageserver_command', None) == SONADOR_COMMAND_IMAGE:
		exitcode = validate_imageserver_image_args(args, exitcode)
	elif getattr(args, 'imageserver_command', None) == SONADOR_COMMAND_PATIENT:
		exitcode = validate_imageserver_patient_args(args, exitcode)
	elif getattr(args, 'imageserver_command', None) == SONADOR_COMMAND_STUDY:
		exitcode = validate_imageserver_study_args(args, exitcode)
	elif getattr(args, 'imageserver_command', None) == SONADOR_COMMAND_JOB:
		exitcode = validate_imageserver_job_args(args, exitcode)

	return exitcode


def validate_imageserver_dicom_modality_args(args, exitcode):
	'''	Ensure that user provided arguments for the Imaging Server DICOM Modality API are sane.

		@input args: arguments to be validated
	'''
	if not getattr(args, 'imageserver_dicom_command', None):
		logger.error('Please specify a DICOM modality command to execute. Supported: %s'
			% ', '.join(SONADOR_COMMAND_SUPPORTED_MODALITY))
		exitcode = 1

	return exitcode


def validate_imageserver_dicomweb_args(args, exitcode):
	'''	Ensure that user provided arguments for the DICOMweb API are sane.

		@input args: arguments to be validated
	'''
	if not getattr(args, 'imageserver_dicomweb_command', None):
		logger.error('Please specify a remote DICOMweb command to execute. Supported: %s'
			% ', '.join(SONADOR_COMMAND_SUPPORTED_DICOMWEB))
		exitcode = 1

	if getattr(args, 'imageserver_dicomweb_command', None) in (SONADOR_COMMAND_REMOTE,):

		# Ensure that a remote server argument has been provided
		if not getattr(args, 'remote_server', None):
			logger.error('Invalid DICOMweb remote server, please provide a valid remote server ID. See --help '
				+ 'for additional information.')
			exitcode = 1

		# Ensure that a remote server interface is specified
		if not getattr(args, 'dicomweb_remote_command', None) :
			logger.error('Please specify a DICOMweb remote management command to execute. Supported: %s'
				% ', '.join(SONADOR_COMMAND_DICOMWEB_REMOTE_SUPPORTED))
			exitcode = 1

		# Validate options for specific interfaces
		if getattr(args, 'dicomweb_remote_command', None) == SONADOR_COMMAND_STUDY:
			exitcode = validate_imageserver_study_args(args, exitcode, 
				studycmd_attr='dicomweb_remote_study_command', supported=SONADOR_COMMAND_DICOMWEB_REMOTECMD_SUPPORTED)
		elif getattr(args, 'dicomweb_remote_command', None) == SONADOR_COMMAND_IMAGE:
			exitcode = validate_imageserver_image_args(args, exitcode,
				imagecmd_attr='dicomweb_remote_images_command', supported=(SONADOR_COMMAND_QUERY, SONADOR_COMMAND_QUERY_FETCH))

	return exitcode


def validate_imageserver_patient_args(args, exitcode, 
		patientcmd_attr='imageserver_patients_command', supported=SONADOR_COMMAND_PATIENT_SUPPORTED):
	'''	Ensure that the user provided arguments for the patient management API are sane

		@input args: arguments to be validated
	'''
	if not getattr(args, patientcmd_attr, None):
		logger.error('Please specify a patient management command to execute. Supported: %s'
			% ', '.join(SONADOR_COMMAND_PATIENT_SUPPORTED))
		exitcode = 1

	return exitcode


def validate_imageserver_study_args(args, exitcode, 
		studycmd_attr='imageserver_study_command', supported=SONADOR_COMMAND_STUDY_SUPPORTED):
	'''	Ensure that user provided arguments for the study API are sane

		@input args: arguments to be validated
	'''	
	if not getattr(args, studycmd_attr, None):
		logger.error('Please specify a study management command to execute. Supported: %s'
			% ', '.join(supported))
		exitcode = 1

	return exitcode


def validate_imageserver_image_args(args, exitcode, 
		imagecmd_attr='imageserver_images_command', supported=SONADOR_COMMAND_SUPPORTED_IMAGE):
	'''	Ensure that the user provided arguments for the image management API are sane.

		@input args: arguments to be validated
	'''
	if not getattr(args, imagecmd_attr, None):
		logger.error('Please specify an image management command to execute. Supported: %s'
			% ', '.join(supported))
		exitcode = 1

	return exitcode


def validate_imageserver_job_args(args, exitcode):
	'''	Ensure that the user provided arguments for the job management API are sane.
	'''
	if not getattr(args, 'imageserver_jobs_command', None):
		logger.error('Please specify an image management command to execute. Supported: %s'
			% ', '.join(SONAOD_COMMAND_SUPPORTED_JOB))
		exitcode = 1

	return exitcode
