import six, os, logging, traceback, argparse, posixpath, json

from client.utils.urls import validate_url
from client.utils.logs import LOGGING_LEVELS

from sonador.apisettings import IMAGING_SERVER_RESOURCE_SERIES, IMAGING_SERVER_RESOURCE_STUDY, IMAGING_SERVER_RESOURCE_PATIENT
from sonador.helpers import SonadorServer, report_operation_error
from sonador.parser import add_arguments_for_api_connection, add_arguments_for_verify_ssl, \
	add_arguments_for_logging_options, output_operation_options, \
	output_destination_options, imageserver_operation_options, datamodel_schema_options, \
	datamodel_query_options, datamodel_query_output_options, datamodel_dicom_modify_options, \
	dicom_download_options, dicomweb_remote_server_operation_options, datamodel_pagination_options, \
	SONADOR_COMMAND_LIST, SONADOR_COMMAND_DETAILS, \
	OUTPUT_TYPE_TABULATE, OUTPUT_TYPE_CSV
from sonador.remote import sonador_datacollection_list, sonador_dataobject_details, sonador_dataobject_schema_display, \
	fetch_sonador_dataobject, sonador_datacollection_serialize, sonador_dataobject_serialize, sonador_dataobject_extendattrs
from sonador.servers import SonadorImagingServerCollection, DicomImagingModalityCollection, RemoteDICOMwebServerCollection, \
	sonador_apitoken_fetch

from cli.apisettings import SONADOR_COMMAND_TOKEN, SONADOR_COMMAND_QUERY, SONADOR_COMMAND_MODIFY, \
	SONADOR_COMMAND_FETCH, SONADOR_COMMAND_QUERY_FETCH, SONADOR_COMMAND_DOWNLOAD, SONADOR_COMMAND_QUERY_DOWNLOAD, \
	SONADOR_COMMAND_REMOVE, SONADOR_COMMAND_QUERY_REMOVE, \
	SONADOR_COMMAND_IMAGE_SERVER, SONADOR_COMMAND_SUPPORTED_IMAGE_SERVER, SONADOR_COMMAND_REMOTE, \
	SONADOR_COMMAND_MODALITY, SONADOR_COMMAND_SUPPORTED_MODALITY, \
	SONADOR_COMMAND_DICOMWEB, SONADOR_COMMAND_SUPPORTED_DICOMWEB, SONADOR_COMMAND_DICOMWEB_REMOTECMD_SUPPORTED, \
	SONADOR_COMMAND_PATIENT, SONADOR_COMMAND_PATIENT_SUPPORTED, \
	SONADOR_COMMAND_STUDY, SONADOR_COMMAND_STUDY_SUPPORTED, \
	SONADOR_COMMAND_IMAGE, SONADOR_COMMAND_IMAGE_UPLOAD, SONADOR_COMMAND_SUPPORTED_IMAGE,  \
	SONADOR_COMMAND_DICOMWEB_REMOTE_SUPPORTED, \
	SONADOR_COMMAND_JOB, SONADOR_COMMAND_PAUSE, SONADOR_COMMAND_RESUME, SONADOR_COMMAND_CANCEL, \
	SONADOR_COMMAND_RESUBMIT, SONAOD_COMMAND_SUPPORTED_JOB, \
	SONADOR_SUPPORTED_COMMANDS
from cli.validate import validate_args, argparse_type_directory, argparse_keyval
from cli.servers import imageserver_datacollection_list, imageserver_dataobject_details, \
	imageserver_upload_folder
from cli.images import download_imaging_filearchive


# Setup command line argument parser
parser = argparse.ArgumentParser(description='Sonador API Client.')
sonador_commands = parser.add_subparsers(dest='command', metavar='cmd',
	title='commands', help='Sonador API command: %s' % ', '.join(SONADOR_SUPPORTED_COMMANDS))

# Connection: Sonador API
add_arguments_for_api_connection(parser)
add_arguments_for_verify_ssl(parser)
parser.add_argument('--internal-dns', dest='internal_dns', action='store_true', default=False,
	help='Use internal DNS (if present) for Sonador and Orthanc API server connections')

# Logging options
add_arguments_for_logging_options(parser)


# Image Server Commands
imageserver_interface = sonador_commands.add_parser(SONADOR_COMMAND_IMAGE_SERVER, 
	help='PACS Imaging Servers API commands')
imageserver_commands = imageserver_interface.add_subparsers(dest='imageserver_command', title='PACS Imaging Server commands',
	metavar='cmd', help='Sonador Image Server Commands. Options: %s' % ', '.join(SONADOR_COMMAND_SUPPORTED_IMAGE_SERVER))

imageserver_list = imageserver_commands.add_parser(SONADOR_COMMAND_LIST, 
	help='List imaging servers currently registered with Sonador')
output_operation_options(imageserver_list)

imageserver_details = imageserver_commands.add_parser(SONADOR_COMMAND_DETAILS,
	help='Retrieve details (expanded data listing) for specific imaging servers registered with Sonador.')
imageserver_details.add_argument('--servers', '-S', dest='servers', nargs='+', type=str, required=True)
datamodel_schema_options(imageserver_details)
output_destination_options(imageserver_details)

imageserver_dicom_modality_interface = imageserver_commands.add_parser(SONADOR_COMMAND_MODALITY, 
	help='PACS DICOM Modality API commands')
imageserver_operation_options(imageserver_dicom_modality_interface)

dicom_modality_commands = imageserver_dicom_modality_interface.add_subparsers(dest='imageserver_dicom_command', 
	title='PACS DICOM Modality commands', metavar='cmd', 
	help='PACS DICOM Modality Commands. Options: %s' % ', '.join(SONADOR_COMMAND_SUPPORTED_MODALITY))

dicom_modality_list = dicom_modality_commands.add_parser(SONADOR_COMMAND_LIST,
	help='List DICOM modalities associated with the server')
output_operation_options(dicom_modality_list)

dicom_modality_details = dicom_modality_commands.add_parser(SONADOR_COMMAND_DETAILS,
	help='Retrieve details (expanded data listing) for specific DICOM modalities associated with an imaging server.')
dicom_modality_details.add_argument('--modalities', '-S', dest='modalities', nargs='+', type=str, required=True)
datamodel_schema_options(dicom_modality_details)
output_destination_options(dicom_modality_details)


# DICOMweb Management Commands
dicomweb_interface = imageserver_commands.add_parser(SONADOR_COMMAND_DICOMWEB,
	help='PACS Remote DICOMWeb API commands')
imageserver_operation_options(dicomweb_interface)

dicomweb_commands = dicomweb_interface.add_subparsers(dest='imageserver_dicomweb_command',
	title='PACS Remote DICOMweb commands', metavar='cmd',
	help='PACS Remote DICOMweb Commands. Options: %s' % ', '.join(SONADOR_COMMAND_SUPPORTED_DICOMWEB))

dicomweb_list = dicomweb_commands.add_parser(SONADOR_COMMAND_LIST,
	help='List remote DICOMweb instances associated with the server')
output_operation_options(dicomweb_list)

dicomweb_details = dicomweb_commands.add_parser(SONADOR_COMMAND_DETAILS,
	help='Retrieve details (expanded data listing) for specific DICOMweb remote servers associated with an imaging server.')
dicomweb_details.add_argument('--remote-servers', '-S', dest='dicomweb_remotes', nargs='+', type=str, required=True)
datamodel_schema_options(dicomweb_details)
output_destination_options(dicomweb_details)


# DICOMWeb Remote Management Commands
dicomweb_remote_interface = dicomweb_commands.add_parser(SONADOR_COMMAND_REMOTE,
	help='DICOMWeb Remote Server commands')
dicomweb_remote_server_operation_options(dicomweb_remote_interface)

dicomweb_remote_commands = dicomweb_remote_interface.add_subparsers(dest='dicomweb_remote_command',
	title='DICOMWeb Remote API commands', metavar='cmd',
	help='DICOMWeb Remote Management Commands. Options: %s' % ', '.join(SONADOR_COMMAND_DICOMWEB_REMOTE_SUPPORTED))

# DICOMWeb Remote Study Management Commands
dicomweb_remote_study_interface = dicomweb_remote_commands.add_parser(SONADOR_COMMAND_STUDY,
	help='DICOMWeb Remote Server Study API commands')
dicomweb_remote_study_commands = dicomweb_remote_study_interface.add_subparsers(dest='dicomweb_remote_study_command',
	title='DICOMWeb Remote Study Management commands', metavar='cmd',
	help='DICOMWeb Remote Study Management Commands. Options: %s' % ', '.join(SONADOR_COMMAND_DICOMWEB_REMOTECMD_SUPPORTED))

dicomweb_remote_study_query = dicomweb_remote_study_commands.add_parser(SONADOR_COMMAND_QUERY,
	help='Query study data stored on the DICOMWeb remote')
datamodel_query_options(dicomweb_remote_study_query)
datamodel_pagination_options(dicomweb_remote_study_query, limit_default=100)
datamodel_query_output_options(dicomweb_remote_study_query)
output_destination_options(dicomweb_remote_study_query)

dicomweb_remote_study_fetch = dicomweb_remote_study_commands.add_parser(SONADOR_COMMAND_FETCH,
	help='Copy studies saved on the DICOMweb remote to the PACS server.')
dicomweb_remote_study_fetch.add_argument('--studies', '-Y', dest='studies', nargs='+', type=str, required=True,
	help='UIDs of studies to be transferred to the PACS server.')
dicomweb_remote_study_fetch.add_argument('--wait', dest='synchronous', action='store_true', default=False,
	help='When true, causes the script to wait until the server returns the result of the job. '
		+ 'Note: There is a risk of connection timeout if the requested studies are large or if there are a large number '
		+ 'of requested resources.')

dicomweb_remote_study_query_fetch = dicomweb_remote_study_commands.add_parser(SONADOR_COMMAND_QUERY_FETCH,
	help='Copy studies saved on the DICOMweb remote to the PACS server which match the provided query.')
datamodel_query_options(dicomweb_remote_study_query_fetch)
datamodel_pagination_options(dicomweb_remote_study_query_fetch, limit_default=100)
datamodel_query_output_options(dicomweb_remote_study_query_fetch)
dicomweb_remote_study_query_fetch.add_argument('--wait', dest='synchronous', action='store_true', default=False,
	help='When true, causes the script to wait until the server returns the result of the job. '
		+ 'Note: There is a risk of connection timeout if the requested resources include a large number of images.')

# DICOMWeb Remote Series Management Commands
dicomweb_remote_images_interface = dicomweb_remote_commands.add_parser(SONADOR_COMMAND_IMAGE,
	help='DICOMweb Remote Server Series/Sequence/Scan API commands')
dicomweb_remote_images_commands = dicomweb_remote_images_interface.add_subparsers(dest='dicomweb_remote_images_command',
	title='DICOMWeb Remote Series/Sequence/Scan Management commands', metavar='cmd',
	help='DICOMWeb Remote Series/Sequence/Scan Management Commands. Options: %s' % ', '.join(SONADOR_COMMAND_DICOMWEB_REMOTECMD_SUPPORTED))

dicomweb_remote_images_query = dicomweb_remote_images_commands.add_parser(SONADOR_COMMAND_QUERY,
	help='Query image sequences/series/scans stored on the DICOMWeb remote')
datamodel_query_options(dicomweb_remote_images_query)
datamodel_pagination_options(dicomweb_remote_images_query, limit_default=100)
datamodel_query_output_options(dicomweb_remote_images_query)
output_destination_options(dicomweb_remote_images_query)

dicomweb_remote_images_query_fetch = dicomweb_remote_images_commands.add_parser(SONADOR_COMMAND_QUERY_FETCH,
	help='Copy sequences/series/scans saved on the DICOMweb remote to the PACS server which match the provided query.')
datamodel_query_options(dicomweb_remote_images_query_fetch)
datamodel_pagination_options(dicomweb_remote_images_query_fetch, limit_default=100)
datamodel_query_output_options(dicomweb_remote_images_query_fetch)
dicomweb_remote_images_query_fetch.add_argument('--wait', dest='synchronous', action='store_true', default=False,
	help='When true, causes the script to wait until the server returns the result of the job. '
		+ 'Note: There is a risk of connection timeout if the requested resources include a large number of images.')


# Patient Management Commands
imageserver_patients_interface = imageserver_commands.add_parser(SONADOR_COMMAND_PATIENT,
	help='PACS Patient API commands')
imageserver_operation_options(imageserver_patients_interface)

patients_commands = imageserver_patients_interface.add_subparsers(dest='imageserver_patients_command',
	title='PACS Patient Management commands', metavar='cmd',
	help='PACS Patient Management Commands. Options: %s' % ', '.join(SONADOR_COMMAND_PATIENT_SUPPORTED))

patient_query = patients_commands.add_parser(SONADOR_COMMAND_QUERY, 
	help='Query patient data stored on the PACS server')
datamodel_query_options(patient_query)
datamodel_pagination_options(patient_query, limit_default=100)
datamodel_query_output_options(patient_query)
output_destination_options(patient_query)

patient_details = patients_commands.add_parser(SONADOR_COMMAND_DETAILS, help='Retrieve patient details')
patient_details.add_argument('--patients', '-P', dest='patients', nargs='+', type=str, required=True,
	help='UIDs of patients for which detailed listings should be retrieved.')
output_destination_options(patient_details)

patient_modify = patients_commands.add_parser(SONADOR_COMMAND_MODIFY, help='Update resource tags and metadata')
patient_modify.add_argument('--patients', '-P', dest='patients', nargs='+', type=str, required=True,
	help='UIDs of the resources for which the updates should be applied.')
datamodel_dicom_modify_options(patient_modify)

patient_download = patients_commands.add_parser(SONADOR_COMMAND_DOWNLOAD, help='Download imaging sequences/studies')
patient_download.add_argument('--patients', '-P', dest='patients', nargs='+', type=str, required=True,
	help='UIDs of the patients for which to download data.')
dicom_download_options(patient_download)

patient_query_download = patients_commands.add_parser(SONADOR_COMMAND_QUERY_DOWNLOAD, 
	help='Query PACS server and download matching patient data.')
datamodel_query_options(patient_query_download)
datamodel_pagination_options(patient_query_download, limit_default=100)
dicom_download_options(patient_query_download)

patient_remove = patients_commands.add_parser(SONADOR_COMMAND_REMOVE, help='Remove patient data')
patient_remove.add_argument('--patients', '-P', dest='patients', nargs='+', type=str, required=True,
	help='UIDS of the patients for which the data should be removed.')

patient_query_remove = patients_commands.add_parser(SONADOR_COMMAND_QUERY_REMOVE, 
	help='Query PACS server and remove matching patient data.')
datamodel_query_options(patient_query_remove)
datamodel_pagination_options(patient_query_remove, limit_default=100)


# Study Management Commands
imageserver_study_interface = imageserver_commands.add_parser(SONADOR_COMMAND_STUDY,
	help='PACS Study API commands')
imageserver_operation_options(imageserver_study_interface)

study_commands = imageserver_study_interface.add_subparsers(dest='imageserver_study_command',
	title='PACS Study Management commands', metavar='cmd',
	help='PACS Study Management Commands. Options: %s' % ', '.join(SONADOR_COMMAND_STUDY_SUPPORTED))

study_query = study_commands.add_parser(SONADOR_COMMAND_QUERY,
	help='Query imaging study data stored on the PACS server')
datamodel_query_options(study_query)
datamodel_pagination_options(study_query, limit_default=1000)
datamodel_query_output_options(study_query)
output_destination_options(study_query)

study_details = study_commands.add_parser(SONADOR_COMMAND_DETAILS, help='Retrieve study details')
study_details.add_argument('--studies', '-Y', dest='studies', nargs='+', type=str, required=True,
	help='UIDs of studies for which detailed listings should be retrieved.')
output_destination_options(study_details)

study_modify = study_commands.add_parser(SONADOR_COMMAND_MODIFY, help='Update resource tags and metadata')
study_modify.add_argument('--studies', '-Y', dest='studies', nargs='+', type=str, required=True,
	help='UIDs of the resources for which the updates should be applied.')
datamodel_dicom_modify_options(study_modify)

study_download = study_commands.add_parser(SONADOR_COMMAND_DOWNLOAD, help='Download imaging studies')
study_download.add_argument('--studies', '-Y', dest='studies', nargs='+', type=str, required=True,
	help='UIDs of the studies to download.')
dicom_download_options(study_download)

study_query_download = study_commands.add_parser(SONADOR_COMMAND_QUERY_DOWNLOAD, 
	help='Query PACS server and download matching studies.')
datamodel_query_options(study_query_download)
datamodel_pagination_options(study_query_download, limit_default=100)
dicom_download_options(study_query_download)

study_remove = study_commands.add_parser(SONADOR_COMMAND_REMOVE, help='Remove imaging studies')
study_remove.add_argument('--studies', '-Y', dest='studies', nargs='+', type=str, required=True,
	help='UIDs of the studies to remove.')

study_query_remove = study_commands.add_parser(SONADOR_COMMAND_QUERY_REMOVE, 
	help='Query PACS server and remove matching study data.')
datamodel_query_options(study_query_remove)
datamodel_pagination_options(study_query_remove, limit_default=100)


# Image Management Commands
imageserver_images_interface = imageserver_commands.add_parser(SONADOR_COMMAND_IMAGE, 
	help='PACS Series/Sequence/Scan API commands')
imageserver_operation_options(imageserver_images_interface)

images_commands = imageserver_images_interface.add_subparsers(dest='imageserver_images_command',
	title='PACS Series/Sequence/Scan Management commands', metavar='cmd',
	help='PACS Series/Sequence/Scan Management Commands. Options: %s' % ', '.join(SONADOR_COMMAND_SUPPORTED_IMAGE))

images_upload = images_commands.add_parser(SONADOR_COMMAND_IMAGE_UPLOAD, 
	help='Upload a directory of DICOM images')
images_upload.add_argument('--folders', '-f', dest='folders', nargs='+', type=argparse_type_directory, required=True,
	help='Folders from which DICOM resources should be sent to the server.')
images_upload.add_argument('--num-workers', dest='threads', default=4, type=int,
	help='Number of background processes that should be used during upload.')

image_query = images_commands.add_parser(SONADOR_COMMAND_QUERY,
	help='Query image sequences/series/scans stored on the PACS server via DICOM tag')
datamodel_query_options(image_query)
datamodel_pagination_options(image_query, limit_default=1000)
datamodel_query_output_options(image_query)
output_destination_options(image_query)

image_details = images_commands.add_parser(SONADOR_COMMAND_DETAILS, help='Retrieve image sequence/series/scan details')
image_details.add_argument('--scans', '-I', dest='scans', nargs='+', type=str, required=True,
	help='UIDs of sequences/series/scans for which detailed listings should be retrieved.')
output_destination_options(image_details)

image_modify = images_commands.add_parser(SONADOR_COMMAND_MODIFY, help='Update resource tags and metadata')
image_modify.add_argument('--scans', '-I', dest='scans', nargs='+', type=str, required=True,
	help='UIDs of the resources for which the updates should be applied.')
datamodel_dicom_modify_options(image_modify)

image_download = images_commands.add_parser(SONADOR_COMMAND_DOWNLOAD, help='Download imaging sequences/series')
image_download.add_argument('--scans', '-I', dest='scans', nargs='+', type=str, required=True,
	help='UIDs of the sequences/series/scans to download')
dicom_download_options(image_download)

image_query_download = images_commands.add_parser(SONADOR_COMMAND_QUERY_DOWNLOAD, 
	help='Query PACS server and download matching images')
datamodel_query_options(image_query_download)
datamodel_pagination_options(image_query_download, limit_default=100)
dicom_download_options(image_query_download)

image_remove = images_commands.add_parser(SONADOR_COMMAND_REMOVE, help='Remove imaging sequences/series')
image_remove.add_argument('--scans', '-I', dest='scans', nargs='+', type=str, required=True,
	help='UIDs of the sequences/series/scans to remove')

image_query_remove = images_commands.add_parser(SONADOR_COMMAND_QUERY_REMOVE, 
	help='Query PACS server and remove matching image data.')
datamodel_query_options(image_query_remove)
datamodel_pagination_options(image_query_remove, limit_default=100)


# Job Management Commands
imageserver_jobs_interface = imageserver_commands.add_parser(SONADOR_COMMAND_JOB,
	help='PACS Job API commands')
imageserver_operation_options(imageserver_jobs_interface)

jobs_commands = imageserver_jobs_interface.add_subparsers(dest='imageserver_jobs_command',
	title='Job Management commands', metavar='cmd',
	help='PACS Job Management Commands. Options: %s' % ', '.join(SONAOD_COMMAND_SUPPORTED_JOB))

jobs_list = jobs_commands.add_parser(SONADOR_COMMAND_LIST,
	help='List processing jobs associated with the server')
output_destination_options(jobs_list)
datamodel_pagination_options(jobs_list, limit_default=1000)

job_details = jobs_commands.add_parser(SONADOR_COMMAND_DETAILS,
	help='Retrieve details (expanded data listing) for specific processing jobs.')
job_details.add_argument('--jobs', '-J', dest='jobs', nargs='+', type=str, required=True,
	help='Job UIDs for which detailed listings should be retrieved.')
output_destination_options(job_details)

job_pause = jobs_commands.add_parser(SONADOR_COMMAND_PAUSE, help='Pause execution of a processing job.')
job_pause.add_argument('--jobs', '-J', dest='jobs', nargs='+', type=str, required=True,
	help='UIDs of jobs which should be paused.')

job_resume = jobs_commands.add_parser(SONADOR_COMMAND_RESUME, help='Resume execution of a processing job.')
job_resume.add_argument('--jobs', '-J', dest='jobs', nargs='+', type=str, required=True,
	help='UIDs of jobs to be resumed.')

job_cancel = jobs_commands.add_parser(SONADOR_COMMAND_CANCEL, help='Cancel execution of a processing job.')
job_cancel.add_argument('--jobs', '-J', dest='jobs', nargs='+', type=str, required=True,
	help='UIDs of jobs to cancel.')

job_resubmit = jobs_commands.add_parser(SONADOR_COMMAND_RESUBMIT, help='Resubmit and re-execute a processing job.')
job_resubmit.add_argument('--jobs', '-J', dest='jobs', nargs='+', type=str, required=True,
	help='UIDs of jobs to resubmit and re-execute.')


# API/Session Management
sessiontoken_fetch = sonador_commands.add_parser(SONADOR_COMMAND_TOKEN, help='Fetch Sonador session token')
output_destination_options(sessiontoken_fetch)



# Parse CLI arguments
args = parser.parse_args()

if args:

	# Configure logging for the script
	logger = logging.getLogger(args.logname)
	logging.basicConfig(level=LOGGING_LEVELS.get(args.loglevel), format=args.logformat)

	# Validate arguments
	try:
		exitcode = validate_args(args)

	except Exception as err:

		# Log any exceptions to output
		logger.error(six.text_type('Unable to execute Sonador API operation. Error: %s.\n' % err))
		if args.error_traceback:
			logger.error(six.text_type('Traceback: %s\n\n%s' % (err, traceback.format_exc())))

		exitcode = 1

	finally:

		# Exit if an error was found during initialization
		if exitcode: exit(exitcode)


# Execute script logic
if __name__ == '__main__':
	exitcode = 0

	try:

		# Sonaodr Connection
		sonador_server = SonadorServer(args.endpoint, args.accessid, args.secretkey, apitoken=args.apitoken, verify=args.verifyssl,
			internal_dns=args.internal_dns)

		# Fetch Sonador Session token
		if args.command == SONADOR_COMMAND_TOKEN:
			logger.info('Fetch session token from server: %s' % args.endpoint)
			sonador_apitoken_fetch(sonador_server, args.output_dest, verify=args.verifyssl)
			args.output_dest.write('\n')

		# Image server commands
		elif args.command == SONADOR_COMMAND_IMAGE_SERVER:

			# List Sonador Imaging Servers
			if args.imageserver_command == SONADOR_COMMAND_LIST:
				filters = {}
				sonador_datacollection_list(sonador_server, args.output_dest, SonadorImagingServerCollection,
					output_type=args.output_type, verify=args.verifyssl, filters=filters)
				args.output_dest.write('\n')

			# Fetch details for Sonador Imaging servers
			elif args.imageserver_command == SONADOR_COMMAND_DETAILS:

				if args.include_schema:
					sonador_dataobject_schema_display(
						sonador_server, args.output_dest, SonadorImagingServerCollection.model, verify=args.verifyssl)

				for iserver_name in args.servers:
					logger.info('Details for Server: %s' % iserver_name)
					iserver = sonador_dataobject_details(sonador_server, args.output_dest, SonadorImagingServerCollection.model, iserver_name,
						verify=args.verifyssl)
					args.output_dest.write('Server URL: %s' % iserver.orthanc_apiurl(''))
					args.output_dest.write('\n')

			
			# DICOM Modality Commands
			elif args.imageserver_command == SONADOR_COMMAND_MODALITY:

				# List DICOM Imaging Modalities for a server
				if args.imageserver_dicom_command == SONADOR_COMMAND_LIST:					
					imageserver_datacollection_list(sonador_server, args, 'dicom', 
						DicomImagingModalityCollection, verify=args.verifyssl)

				# Fetch details for DICOM Modalities
				elif args.imageserver_dicom_command == SONADOR_COMMAND_DETAILS:

					imageserver_dataobject_details(sonador_server, args, 'dicom', 
						DicomImagingModalityCollection.model, args.modalities, 'Details for Server %s Modality %s',
						verify=args.verifyssl)

			
			# DICOMWeb Commands
			elif args.imageserver_command == SONADOR_COMMAND_DICOMWEB:

				# List remote DICOMweb instances	
				if args.imageserver_dicomweb_command == SONADOR_COMMAND_LIST:
					
					imageserver_datacollection_list(sonador_server, args, 'dicom-web',
						RemoteDICOMwebServerCollection, verify=args.verifyssl)

				# Fetch details for DICOMweb Remotes
				elif args.imageserver_dicomweb_command == SONADOR_COMMAND_DETAILS:
					
					imageserver_dataobject_details(sonador_server, args, 'dicom-web',
						RemoteDICOMwebServerCollection.model, args.dicomweb_remotes, 'Details for Server %s DICOMweb Remote %s',
						verify=args.verifyssl)

				# DICOMweb Remote Commands
				elif args.imageserver_dicomweb_command == SONADOR_COMMAND_REMOTE:

					# Retrieve PACS imaging server and remote DICOMweb instance
					rdweb = fetch_sonador_dataobject(
							sonador_server, SonadorImagingServerCollection.model, args.server, verify=args.verifyssl) \
						.get_dicomweb_remote(args.remote_server)

					logger.debug('Configuration for DICOMweb remote %s for PACS %s retrieved successfully'
						% (rdweb.pk, rdweb.pacs.server_label))

					# Execute commands against remote study API
					if args.dicomweb_remote_command == SONADOR_COMMAND_STUDY:
						logger.debug('Execute study API command for remote DICOMweb %s' % rdweb.pk)

						# Execute remote query
						if args.dicomweb_remote_study_command == SONADOR_COMMAND_QUERY:
							results = rdweb.remote_query(dict((k,v) for k,v in args.query), 
								expand=not args.collapse, resource=IMAGING_SERVER_RESOURCE_STUDY,
								limit=args.limit, offset=args.offset)

							if args.collapse:
								args.output_dest.write(' '.join(results))
							else:
								sonador_datacollection_serialize(results, args.output_dest)
								args.output_dest.write('\n')

						# Fetch remote studies
						elif args.dicomweb_remote_study_command == SONADOR_COMMAND_FETCH:
							j = rdweb.remote_fetch([{ 'Study': sid } for sid in args.studies], async_transfer=not args.synchronous)

							if args.synchronous: logger.info('Studies transferred successfully.')
							else: logger.info('Resource request submitted successfully. Job ID: %s.' % j.pk)

						# Query/fetch remote studies
						elif args.dicomweb_remote_study_command == SONADOR_COMMAND_QUERY_FETCH:
							results = rdweb.remote_query(dict((k,v) for k,v in args.query), 
								resource=IMAGING_SERVER_RESOURCE_STUDY,
								limit=args.limit, offset=args.offset)
							logger.info('%s results found which matched query' % len(results))

							# Submit transfer job for results
							if len(results):
								j = rdweb.remote_fetch([{ 'Study': r.pk, } for r in results],
									async_transfer=not args.synchronous)

								if args.synchronous:
									logger.info('%s studies transferred successfully.' % len(results))
								else:
									logger.info('Transfer request submitted successfully for %s resources. Job ID: %s.'
										% (len(results), j.pk))


					# Execute commands against remote series/sequence/scan API
					elif args.dicomweb_remote_command == SONADOR_COMMAND_IMAGE:
						logger.debug('Execute series/sequence/scan API command for remote DICOMweb %s' % rdweb.pk)

						# Execute remote query
						if args.dicomweb_remote_images_command == SONADOR_COMMAND_QUERY:
							results = rdweb.remote_query(dict((k,v) for k,v in args.query), 
								expand=not args.collapse, resource=IMAGING_SERVER_RESOURCE_SERIES,
								limit=args.limit, offset=args.offset)

							if args.collapse:
								args.output_dest.write(' '.join(results))
							else:
								sonador_datacollection_serialize(results, args.output_dest)
								args.output_dest.write('\n')

						# Query/fetch remote series
						elif args.dicomweb_remote_images_command == SONADOR_COMMAND_QUERY_FETCH:
							results = rdweb.remote_query(dict((k,v) for k,v in args.query), 
								resource=IMAGING_SERVER_RESOURCE_SERIES,
								limit=args.limit, offset=args.offset)
							logger.info('%s results found which matched query' % len(results))

							# Submit transfer job for results
							if len(results):
								j = rdweb.remote_fetch([{ 'Study': r.study, 'Series': r.pk } for r in results],
									async_transfer=not args.synchronous)

								if args.synchronous:
									logger.info('%s series/scans/sequences transferred successfully.' % len(results))
								else:
									logger.info('Transfer request submitted successfully for %s resources. Job ID: %s.'
										% (len(results), j.pk))

			
			# Patient Commands
			elif args.imageserver_command == SONADOR_COMMAND_PATIENT:

				# Retrieve PACS imaging server details
				iserver = fetch_sonador_dataobject(
					sonador_server, SonadorImagingServerCollection.model, args.server, verify=args.verifyssl)

				# Query the server for patients
				if args.imageserver_patients_command == SONADOR_COMMAND_QUERY:

					# Execute resource query
					results = iserver.query(dict((k,v) for k,v in args.query), expand=not args.collapse, resource=IMAGING_SERVER_RESOURCE_PATIENT,
						limit=args.limit, offset=args.offset)

					# Output space separate list of patient IDs
					if args.collapse:
						args.output_dest.write(' '.join(results))
					else:
						sonador_datacollection_serialize(results, args.output_dest)
						args.output_dest.write('\n')

				# Retrieve patient details
				elif args.imageserver_patients_command == SONADOR_COMMAND_DETAILS:

					# Retrieve patient details
					for pid in args.patients:

						logger.debug('Retrieve patient %s from server %s' % (pid, iserver.server_label))
						p = iserver.get_patient(pid)
						logger.info('Details for Patient %s on Server %s' % (p.pk, iserver.server_label))
						sonador_dataobject_serialize(p, args.output_dest, include_extended_attrs=False)
						if p.studies:
							logger.info('--- Patient Studies ---')
							args.output_dest.write('\n'.join(p.studies))
							args.output_dest.write('\n')
						
						# Raw object attributes and metadata
						logger.debug('Core Data:\n%s' % json.dumps(p._objectdata))
						logger.debug('Patient Meta:\n%s' % json.dumps(p.meta))
						args.output_dest.write('\n')

				# Modify image DICOM tags on studies
				elif args.imageserver_patients_command == SONADOR_COMMAND_MODIFY:

					# Make requested modifications to studies
					for pid in args.patients:
						logger.debug('Retrieve patient %s from server %s' % (pid, iserver.server_label))
						p = iserver.get_patient(pid)
						p.modify(replace=dict((k,v) for k,v in args.modify), remove=args.remove_tags,
							private_creator=args.private_creator)
						logger.info('DICOM tags/metadata for patient %s on server %s updated successfully' 
							% (pid, iserver.pk))

				# Download patient DICOM data
				elif args.imageserver_patients_command == SONADOR_COMMAND_DOWNLOAD:

					# Download requested scan/series
					for pid in args.patients:
						logger.debug('Retrieve DICOM data for patient %s from server %s' % (pid, iserver.server_label))
						
						# Retrieve DICOM data from PACS server
						p = iserver.get_patient(pid)
						download_imaging_filearchive(p, args.download_folder, extract=args.extract)

				# Query server and download patient DICOM data
				elif args.imageserver_patients_command == SONADOR_COMMAND_QUERY_DOWNLOAD:

					# Execute resource query
					results = iserver.query(dict((k,v) for k,v in args.query), resource=IMAGING_SERVER_RESOURCE_PATIENT,
						limit=args.limit, offset=args.offset)
					logger.info('Query results: %s matches' % len(results))

					# Retrieve study DICOM data from PACS server
					for p in results:
						download_imaging_filearchive(p, args.download_folder, extract=args.extract)

				# Remove patient DICOM data
				elif args.imageserver_patients_command == SONADOR_COMMAND_REMOVE:

					# Retrieve reference to image and remove
					for pid in args.patients:
						logger.debug('Retrieve DICOM data for patient %s from server %s' % (pid, iserver.server_label))
						p = iserver.get_patient(pid)
						p.delete()
						logger.info('DICOM data for patient %s removed from server %s successfully' % (pid, iserver.server_label))

				# Query server and remove patient DICOM data
				elif args.imageserver_patients_command == SONADOR_COMMAND_QUERY_REMOVE:

					# Execute resource query
					results = iserver.query(dict((k,v) for k,v in args.query), resource=IMAGING_SERVER_RESOURCE_PATIENT,
						limit=args.limit, offset=args.offset)
					logger.info('Query results: %s matches' % len(results))

					# Retrieve reference to image and remove
					for r in results:
						r.delete()
						logger.info('DICOM data for patient %s removed from server %s successfully' % (r.pk, iserver.server_label))

			
			# Study Commands
			elif args.imageserver_command == SONADOR_COMMAND_STUDY:

				# Retrieve PACS imaging server details
				iserver = fetch_sonador_dataobject(
					sonador_server, SonadorImagingServerCollection.model, args.server, verify=args.verifyssl)

				# Query the server for studies
				if args.imageserver_study_command == SONADOR_COMMAND_QUERY:

					# Execute resource query
					results = iserver.query(dict((k,v) for k,v in args.query), expand=not args.collapse, resource=IMAGING_SERVER_RESOURCE_STUDY,
						limit=args.limit, offset=args.offset)

					# Output space separate list of patient IDs
					if args.collapse:
						args.output_dest.write(' '.join(results))
					else:
						sonador_datacollection_serialize(results, args.output_dest)
						args.output_dest.write('\n')

				# Retrieve study details
				elif args.imageserver_study_command == SONADOR_COMMAND_DETAILS:

					for sid in args.studies:

						logger.debug('Retrieve study %s from server %s' % (sid, iserver.server_label))
						s = iserver.get_study(sid)
						logger.info('Details for Study %s on Server %s' % (s.pk, iserver.server_label))
						sonador_dataobject_serialize(s, args.output_dest, include_extended_attrs=False)
						if s.patientdata:
							logger.info('--- Patient Data ---')
							sonador_dataobject_extendattrs(s.patientdata, args.output_dest, six.iterkeys(s.patientdata))
						if s.dicomdata:
							logger.info('--- Study DICOM ---')
							sonador_dataobject_extendattrs(s.dicomdata, args.output_dest, six.iterkeys(s.dicomdata))
						if s.series:
							logger.info('--- Instance Series UIDs (%s Series) ---' % len(s.series))
							args.output_dest.write('\n'.join(s.series))
							args.output_dest.write('\n')

						# Raw object attributes and metadata
						logger.debug('Core Data:\n%s' % json.dumps(s._objectdata))
						logger.debug('Study Meta:\n%s' % json.dumps(s.meta))
						args.output_dest.write('\n')

				# Modify image DICOM tags on studies
				elif args.imageserver_study_command == SONADOR_COMMAND_MODIFY:

					# Make requested modifications to studies
					for sid in args.studies:
						logger.debug('Retrieve study %s from server %s' % (sid, iserver.server_label))
						s = iserver.get_study(sid)
						s.modify(replace=dict((k,v) for k,v in args.modify), remove=args.remove_tags,
							private_creator=args.private_creator)
						logger.info('DICOM tags/metadata for study %s on server %s updated successfully' 
							% (sid, iserver.pk))

				# Download study DICOM data
				elif args.imageserver_study_command == SONADOR_COMMAND_DOWNLOAD:

					# Download requested scan/series
					for sid in args.studies:
						logger.debug('Retrieve DICOM data for study %s from server %s' % (sid, iserver.server_label))
						
						# Retrieve DICOM data from PACS server
						s = iserver.get_study(sid)
						download_imaging_filearchive(s, args.download_folder, extract=args.extract)

				# Query server and download study DICOM data
				elif args.imageserver_study_command == SONADOR_COMMAND_QUERY_DOWNLOAD:

					# Execute resource query
					results = iserver.query(dict((k,v) for k,v in args.query), resource=IMAGING_SERVER_RESOURCE_STUDY,
						limit=args.limit, offset=args.offset)
					logger.info('Query results: %s matches' % len(results))

					# Retrieve study DICOM data from PACS server
					for s in results:
						download_imaging_filearchive(s, args.download_folder, extract=args.extract)

				# Remove study DICOM data
				elif args.imageserver_study_command == SONADOR_COMMAND_REMOVE:

					# Retrieve reference to image and remove
					for sid in args.studies:
						logger.debug('Retrieve DICOM data for study %s from server %s' % (sid, iserver.server_label))
						s = iserver.get_study(sid)
						s.delete()
						logger.info('DICOM data for study %s removed from server %s successfully' % (sid, iserver.server_label))

				# Query server and remove study DICOM data
				elif args.imageserver_study_command == SONADOR_COMMAND_QUERY_REMOVE:

					# Execute resource query
					results = iserver.query(dict((k,v) for k,v in args.query), resource=IMAGING_SERVER_RESOURCE_STUDY,
						limit=args.limit, offset=args.offset)
					logger.info('Query results: %s matches' % len(results))

					# Retrieve reference to image and remove
					for r in results:
						r.delete()
						logger.info('DICOM data for study %s removed from server %s successfully' % (r.pk, iserver.server_label))
			
			# PACS Image Commands
			elif args.imageserver_command == SONADOR_COMMAND_IMAGE:

				# Retrieve PACS imaging server details
				iserver = fetch_sonador_dataobject(
					sonador_server, SonadorImagingServerCollection.model, args.server, verify=args.verifyssl)

				# Upload images to Orthanc
				if args.imageserver_images_command == SONADOR_COMMAND_IMAGE_UPLOAD:
					imageserver_upload_folder(iserver, args.folders, verify=args.verifyssl, threads=args.threads)

				# Query the server for images
				elif args.imageserver_images_command == SONADOR_COMMAND_QUERY:

					# Execute resource query
					results = iserver.query(dict((k,v) for k,v in args.query), expand=not args.collapse, resource=IMAGING_SERVER_RESOURCE_SERIES, 
						limit=args.limit, offset=args.offset)
					
					# Output space separated list of study IDs
					if args.collapse:
						args.output_dest.write(' '.join(results))
					else:
						sonador_datacollection_serialize(results, args.output_dest)
						args.output_dest.write('\n')

				# Retrieve image scan/series detais
				elif args.imageserver_images_command == SONADOR_COMMAND_DETAILS:

					# Retrieve details of scans/series
					for rid in args.scans:

						logger.debug('Retrieve series %s from server %s' % (rid, iserver.server_label))
						i = iserver.get_series(rid)
						logger.info('Details for Scan/Series %s on Server %s' % (i.pk, iserver.server_label))
						sonador_dataobject_serialize(i, args.output_dest, include_extended_attrs=False)
						if i.patientdata:
							logger.info('--- Patient DICOM ---')
							sonador_dataobject_extendattrs(i.patientdata, args.output_dest, six.iterkeys(i.patientdata))
						if i.dicomdata:
							logger.info('--- Series DICOM ---')
							sonador_dataobject_extendattrs(i.dicomdata, args.output_dest, six.iterkeys(i.dicomdata))
						if i.slices:
							logger.info('--- Instance UIDs (%s Images) ---' % len(i.slices))
							args.output_dest.write('\n'.join(i.slices))
							args.output_dest.write('\n')
						
						# Raw object attributes and metadata
						logger.debug('Core Data:\n%s' % json.dumps(i._objectdata))
						logger.debug('Series Meta:\n%s' % json.dumps(i.meta))
						args.output_dest.write('\n')

				# Modify image DICOM tags on scans
				elif args.imageserver_images_command == SONADOR_COMMAND_MODIFY:

					# Make requested modifications to scans/series
					for rid in args.scans:
						logger.debug('Retrieve series %s from server %s' % (rid, iserver.server_label))
						i = iserver.get_series(rid)
						i.modify(replace=dict((k,v) for k,v in args.modify), remove=args.remove_tags,
							private_creator=args.private_creator)
						logger.info('DICOM tags/metadata for scan/series %s on server %s updated successfully' 
							% (rid, iserver.pk))

				# Download image DICOM data
				elif args.imageserver_images_command == SONADOR_COMMAND_DOWNLOAD:

					# Download requested scan/series
					for rid in args.scans:
						logger.debug('Retrieve DICOM data for series %s from server %s' % (rid, iserver.server_label))
						
						# Retrieve study DICOM data from PACS server
						i = iserver.get_series(rid)
						download_imaging_filearchive(i, args.download_folder, extract=args.extract)

				# Query server and download study DICOM data
				elif args.imageserver_images_command == SONADOR_COMMAND_QUERY_DOWNLOAD:

					# Execute resource query
					results = iserver.query(dict((k,v) for k,v in args.query), resource=IMAGING_SERVER_RESOURCE_SERIES,
						limit=args.limit, offset=args.offset)
					logger.info('Query results: %s' % len(results))

					# REtrieve study DICOM data from PACS server
					for i in results:
						download_imaging_filearchive(i, args.download_folder, extract=args.extract)

				# Remove image DICOM data
				elif args.imageserver_images_command == SONADOR_COMMAND_REMOVE:

					# Retrieve reference to image and remove
					for rid in args.scans:
						logger.debug('Retrieve DICOM data for series %s from server %s' % (rid, iserver.server_label))
						i = iserver.get_series(rid)
						i.delete()
						logger.info('DICOM data for series %s fremoved from server %s successfully' % (rid, iserver.server_label))

				# Query server and remove image DICOM data
				elif args.imageserver_images_command == SONADOR_COMMAND_QUERY_REMOVE:

					# Execute resource query
					results = iserver.query(dict((k,v) for k,v in args.query), resource=IMAGING_SERVER_RESOURCE_STUDY,
						limit=args.limit, offset=args.offset)
					logger.info('Query results: %s matches' % len(results))

					# Retrieve reference to image and remove
					for r in results:
						r.delete()
						logger.info('DICOM data for series %s removed from server %s successfully' % (r.pk, iserver.server_label))

			# PACS Job Commands
			elif args.imageserver_command == SONADOR_COMMAND_JOB:

				# Retrieve PACS imaging server details
				iserver = fetch_sonador_dataobject(
					sonador_server, SonadorImagingServerCollection.model, args.server, verify=args.verifyssl)

				# List currently active server jobs
				if args.imageserver_jobs_command == SONADOR_COMMAND_LIST:
					jobs = iserver.fetch_jobs(limit=args.limit, offset=args.offset)
					sonador_datacollection_serialize(jobs, args.output_dest)
					args.output_dest.write('\n')

				# Retrieve job details
				elif args.imageserver_jobs_command == SONADOR_COMMAND_DETAILS:

					for jid in args.jobs:

						logger.debug('Retrieve details for job %s from server %s' % (jid, iserver.server_label))
						j = iserver.get_job(jid)
						logger.info('Details for Job %s on Server %s' % (j.pk, iserver.server_label))
						sonador_dataobject_serialize(j, args.output_dest, include_extended_attrs=False)
						if j.content:
							logger.info('--- Details ---')
							sonador_dataobject_extendattrs(j.content, args.output_dest, six.iterkeys(j.content))
						if j.error:
							logger.info('--- Error ---')
							sonador_dataobject_extendattrs(j, args.output_dest, ('error', 'error_description'))

						# Raw object attributes and metadata
						logger.debug('Core Data:\n%s' % json.dumps(j._objectdata))
						args.output_dest.write('\n')

				# Resubmit failed job
				elif args.imageserver_jobs_command == SONADOR_COMMAND_RESUBMIT:

					for jid in args.jobs:
						logger.debug('Retrieve details for job %s from server %s' % (jid, iserver.server_label))

						# Retrieve job
						j = iserver.get_job(jid)
						j.retry()
						logger.info('Job %s resubmitted to server %s successfully.' % (j.pk, iserver.server_label))

				# Pause currently executing jobs
				elif args.imageserver_jobs_command == SONADOR_COMMAND_PAUSE:

					for jid in args.jobs:
						logger.debug('Retrieve details for job %s from server %s' % (jid, iserver.server_label))

						# Retrieve job
						j = iserver.get_job(jid)
						j.pause()
						logger.info('Job %s on server %s paused successfully.' % (j.pk, iserver.server_label))

				# Resume previously paused jobs
				elif args.imageserver_jobs_command == SONADOR_COMMAND_RESUME:

					for jid in args.jobs:
						logger.debug('Retrieve details for job %s from server %s' % (jid, iserver.server_label))

						# Retrieve job
						j = iserver.get_job(jid)
						j.resume()
						logger.info('Job %s on server %s resumed successfully.' % (j.pk, iserver.server_label))

				# Cancel job
				elif args.imageserver_jobs_command == SONADOR_COMMAND_CANCEL:

					for jid in args.jobs:
						logger.debug('Retrieve details for job %s from server %s' % (jid, iserver.server_label))

						# Retrieve job
						j = iserver.get_job(jid)
						j.cancel()
						logger.info('Job %s on server %s canceled successfully.' % (j.pk, iserver.server_label))
	
	
	except Exception as err:
		report_operation_error(err, error_traceback=traceback.format_exc() if args.error_traceback else None)
		exitcode = 1

	
	finally:
		exit(exitcode)
